/********************************************************************************
* QLogic ISP2xxx Firmware Dump Reader
* Copyright (C) 2006 QLogic Corporation
* (www.qlogic.com)
*
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the
* Free Software Foundation; either version 2, or (at your option) any
* later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
******************************************************************************/

/* QLogic Firmware Dump Reader
 * Usage: ./qla_gdm <host_no>. 
 * The host_no can be obtained from /proc/scsi/qla2300/
 * Original author: AV
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdint.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "qla_gdmp.h"
#include "../exioctln.h"

extern UINT32 qlapi_get_host_no_by_handle(UINT32, UINT16 *);

static void
report(const char *prefix, const char *err, va_list params)
{
	fputs(prefix, stderr);
	vfprintf(stderr, err, params);
	fputs("\n", stderr);
}

void
die(const char *err, ...)
{
	va_list params;

	va_start(params, err);
	report("Fatal: ", err, params);
	va_end(params);
	exit(128);
}

int
error(const char *err, ...)
{
	va_list params;

	va_start(params, err);
	report("Error: ", err, params);
	va_end(params);
	return -1;
}

int
info(const char *err, ...)
{
	va_list params;

	va_start(params, err);
	report("Info: ", err, params);
	va_end(params);
	return 0;
}

#define DEFAULT_ALLOC 0x100000

static inline void *
xmalloc(size_t size)
{
	void *ret = malloc(size);
	if (!ret && !size)
		ret = malloc(1);
	if (!ret)
		die("Out of memory, malloc failed");
	return ret;
}

static inline void *
xrealloc(void *ptr, size_t size)
{
	void *ret = realloc(ptr, size);
	if (!ret && !size)
		ret = realloc(ptr, 1);
	if (!ret)
		die("Out of memory, realloc failed");
	return ret;
}

static inline ssize_t
xwrite(int fd, const void *buf, size_t len)
{
	ssize_t nr;
	while (1) {
		nr = write(fd, buf, len);
		if ((nr < 0) && (errno == EAGAIN || errno == EINTR))
			continue;
		return nr;
	}
}

#define PATH_MAX 4096

void
write_dump(int host_no, SD_UINT8 *buf, SD_UINT32 size)
{
	int fd;
	ssize_t written;
	char time[30];
	char path[PATH_MAX];
	struct timeval tv;
	time_t curtime;

	gettimeofday(&tv, NULL);
	curtime=tv.tv_sec;
	strftime(time, 30, "%Y%m%d_%H%M%S", localtime(&curtime));
	snprintf(path, PATH_MAX, "fw_dump_%d_%s.bin", host_no, time);
	info("Firmware dumped to file %s", path);

	fd = open(path, O_WRONLY|O_CREAT|O_TRUNC, 0666);
	if (fd < 0) {
		error("Firmware dump on %d -- can't open %s for writing %d", host_no,
		    path, errno);
		return;
	}
	written = xwrite(fd, buf, size);
	if (written == size)
		info("Firmware dump on %d -- dumped %d bytes...", host_no, size);
	else
		error("Firmware dump on %d -- short write %d of %d bytes...", host_no,
		    written, size);
	close(fd);
}

int
get_dump(SDMGT_HANDLE handle, int host_no)
{
	int ret;
	SD_UINT32 result, osize, size;
	SD_UINT8 *buf;

	ret = 1;
	osize = size = DEFAULT_ALLOC;
	buf = xmalloc(size);
retry:
	result = SDGetFwDump(handle, 0, buf, &size);
	switch (result) {
	case SDMGT_NONE:
		write_dump(host_no, buf, size);
		break;
	case SDMGT_HBA_NOT_READY:
		info("No firmware dump available at host %d", host_no);
		break;
	case SDMGT_BUFFER_TOO_SMALL:
		info("Firmware dump on %d -- expanding buffer %d --> %d", host_no,
		    osize, size);
		osize = size;
		size += DEFAULT_ALLOC;
		buf = realloc(buf, size);
		goto retry;
	default:
		error("Unable to retrieve dump on %d -- (%s)", host_no,
		    SDGetErrorString(result));
		ret = 1;
		break;
	}
	free(buf);
	return ret;
}

int
main(int argc, const char **argv, char **envp)
{
	int host_no, done;
 	SD_UINT16 hno;
       	SD_UINT32 result, inst, size;

	SDMGT_HANDLE handle;

	if (argc < 2) {
		printf("QLogic Firmware Dump Reader\n");
		printf("Usage:\n");
		printf("%s: <host_no>\n", argv[0]);
		exit (128);
	}

	done = 0;
	inst = 0;
	host_no = strtol(argv[1], NULL, 10);
	do {
		result = SDOpenDevice(inst, &handle);
		switch (result) {
		case SDMGT_NONE:
			if (qlapi_get_host_no_by_handle((SD_UINT32)handle,
			    &hno))
				error("No host_no for instance %d", inst);
			if (hno != host_no)
				break;

			get_dump(handle, host_no);
			done = 1;
			break;
		case SDMGT_UNSUPPORTED_VERSION:
			error("Unsupported instance %d", inst);
			break;
		case SDMGT_DEV_NOT_FOUND:
		default:
			error("host_no %d not found", host_no);
			done = 1;
			break;
		}

		/* Close the device */
		if (result == SDMGT_NONE) {
			SDCloseDevice(handle);
		}
		inst++;
	} while (!done);

	return 0;
}
